# -*- coding: utf-8 -*-

""" This module is designed to contain classes that act as simplified / reduced
energy specific interfaces (facades) for solph components to simplify its
application.
"""

__copyright__ = "oemof developer group"
__license__ = "GPLv3"
